import logging
import math

from bdffont import BdfFont, BdfGlyph, xlfd

from pixel_font_builder import font
from pixel_font_builder.info import SerifMode, WidthMode

logger = logging.getLogger('pixel_font_builder.bdf')


class Configs:
    def __init__(
            self,
            resolution_x: int = 75,
            resolution_y: int = 75,
    ):
        self.resolution_x = resolution_x
        self.resolution_y = resolution_y


def _create_glyph(context: 'font.FontBuilder', code_point: int, glyph_name: str) -> BdfGlyph:
    glyph = context.get_glyph(glyph_name)
    scalable_width_x = math.ceil((glyph.advance_width / context.size) * (72 / context.bdf_configs.resolution_x) * 1000)
    return BdfGlyph(
        name=glyph.name,
        code_point=code_point,
        scalable_width=(scalable_width_x, 0),
        device_width=(glyph.advance_width, 0),
        bounding_box_size=glyph.size,
        bounding_box_offset=glyph.offset,
        bitmap=glyph.data,
    )


def create_builder(context: 'font.FontBuilder') -> BdfFont:
    logger.debug("Create 'BDFBuilder': %s", context.meta_infos.family_name)
    context.check_ready()

    builder = BdfFont(
        point_size=context.size,
        resolution_xy=(context.bdf_configs.resolution_x, context.bdf_configs.resolution_y),
        bounding_box_size=(context.size, context.line_height),
        bounding_box_offset=(0, context.descent),
    )

    logger.debug("Add 'Glyph': .notdef")
    builder.add_glyph(_create_glyph(context, -1, '.notdef'))
    for code_point, glyph_name in context.character_mapping.items():
        logger.debug("Add 'Glyph': %s", glyph_name)
        builder.add_glyph(_create_glyph(context, code_point, glyph_name))

    logger.debug("Setup 'Properties'")
    builder.properties.foundry = context.meta_infos.manufacturer
    builder.properties.family_name = context.meta_infos.family_name
    builder.properties.weight_name = context.meta_infos.style_name
    builder.properties.slant = xlfd.Slant.ROMAN
    builder.properties.setwidth_name = xlfd.SetwidthName.NORMAL
    if context.meta_infos.serif_mode == SerifMode.SERIF:
        builder.properties.add_style_name = xlfd.AddStyleName.SERIF
    elif context.meta_infos.serif_mode == SerifMode.SANS_SERIF:
        builder.properties.add_style_name = xlfd.AddStyleName.SANS_SERIF
    else:
        builder.properties.add_style_name = context.meta_infos.serif_mode
    builder.properties.pixel_size = context.size
    builder.properties.point_size = context.size * 10
    builder.properties.resolution_x = context.bdf_configs.resolution_x
    builder.properties.resolution_y = context.bdf_configs.resolution_y
    if context.meta_infos.width_mode == WidthMode.MONOSPACED:
        builder.properties.spacing = xlfd.Spacing.MONOSPACED
    elif context.meta_infos.width_mode == WidthMode.PROPORTIONAL:
        builder.properties.spacing = xlfd.Spacing.PROPORTIONAL
    else:
        builder.properties.spacing = context.meta_infos.width_mode
    builder.properties.average_width = round(sum([glyph.device_width_x * 10 for glyph in builder.code_point_to_glyph.values()]) / builder.get_glyphs_count())
    builder.properties.charset_registry = xlfd.CharsetRegistry.ISO10646
    builder.properties.charset_encoding = '1'

    builder.properties.default_char = -1
    builder.properties.font_ascent = context.ascent
    builder.properties.font_descent = context.descent
    builder.properties.x_height = context.x_height
    builder.properties.cap_height = context.cap_height

    builder.properties.font_version = context.meta_infos.version
    builder.properties.copyright = context.meta_infos.copyright_info
    builder.properties['LICENSE'] = context.meta_infos.license_info

    builder.generate_xlfd_font_name()

    logger.debug("Create 'BDFBuilder' finished")
    return builder



from enum import StrEnum


class StyleName(StrEnum):
    LIGHT = 'Light'
    NORMAL = 'Normal'
    REGULAR = 'Regular'
    MEDIUM = 'Medium'
    BOLD = 'Bold'
    HEAVY = 'Heavy'


class SerifMode(StrEnum):
    SERIF = 'Serif'
    SANS_SERIF = 'Sans-Serif'


class WidthMode(StrEnum):
    MONOSPACED = 'Monospaced'
    PROPORTIONAL = 'Proportional'


class MetaInfos:
    def __init__(
            self,
            version: str = '0.0.0',
            family_name: str = None,
            style_name: str = StyleName.REGULAR,
            serif_mode: SerifMode = None,
            width_mode: WidthMode = None,
            manufacturer: str = None,
            designer: str = None,
            description: str = None,
            copyright_info: str = None,
            license_info: str = None,
            vendor_url: str = None,
            designer_url: str = None,
            license_url: str = None,
            sample_text: str = None,
    ):
        self.version = version
        self.family_name = family_name
        self.style_name = style_name
        self.serif_mode = serif_mode
        self.width_mode = width_mode
        self.manufacturer = manufacturer
        self.designer = designer
        self.description = description
        self.copyright_info = copyright_info
        self.license_info = license_info
        self.vendor_url = vendor_url
        self.designer_url = designer_url
        self.license_url = license_url
        self.sample_text = sample_text




class Glyph:
    def __init__(
            self,
            name: str,
            advance_width: int = 0,
            offset: tuple[int, int] = (0, 0),
            data: list[list[int]] = None,
    ):
        self.name = name
        self.advance_width = advance_width
        self.offset_x, self.offset_y = offset
        if data is None:
            data = list[list[int]]()
        self.data = data

    @property
    def offset(self) -> tuple[int, int]:
        return self.offset_x, self.offset_y

    @offset.setter
    def offset(self, value: tuple[int, int]):
        self.offset_x, self.offset_y = value

    @property
    def width(self) -> int:
        if len(self.data) > 0:
            return len(self.data[0])
        else:
            return 0

    @property
    def height(self) -> int:
        return len(self.data)

    @property
    def size(self) -> tuple[int, int]:
        return self.width, self.height

    def check_ready(self):
        if any(len(data_row) != len(self.data[0]) for data_row in self.data):
            raise Exception(f"Glyph '{self.name}': lengths of data rows are not equals")



import logging
from enum import StrEnum

from fontTools.fontBuilder import FontBuilder
from fontTools.misc.psCharStrings import T2CharString as OTFGlyph
from fontTools.pens.t2CharStringPen import T2CharStringPen as OTFGlyphPen
from fontTools.pens.ttGlyphPen import TTGlyphPen as TTFGlyphPen
from fontTools.ttLib import TTCollection
from fontTools.ttLib.tables._g_l_y_f import Glyph as TTFGlyph

from pixel_font_builder import font
from pixel_font_builder.glyph import Glyph
from pixel_font_builder.info import MetaInfos

logger = logging.getLogger('pixel_font_builder.opentype')

_CACHE_NAME_TAG = '_opentype_cache_tag'
_CACHE_NAME_OUTLINES = '_opentype_cache_outlines'
_CACHE_NAME_OTF_GLYPH = '_opentype_cache_otf_glyph'
_CACHE_NAME_TTF_GLYPH = '_opentype_cache_ttf_glyph'


class Configs:
    def __init__(
            self,
            px_to_units: int = 100,
            cff_family_name: str = None,
    ):
        self.px_to_units = px_to_units
        self.cff_family_name = cff_family_name


class Flavor(StrEnum):
    WOFF = 'woff'
    WOFF2 = 'woff2'


def _create_name_strings(meta_infos: MetaInfos) -> dict[str, str]:
    """
    https://learn.microsoft.com/en-us/typography/opentype/spec/name#name-ids
    copyright (nameID 0)
    familyName (nameID 1)
    styleName (nameID 2)
    uniqueFontIdentifier (nameID 3)
    fullName (nameID 4)
    version (nameID 5)
    psName (nameID 6)
    trademark (nameID 7)
    manufacturer (nameID 8)
    designer (nameID 9)
    description (nameID 10)
    vendorURL (nameID 11)
    designerURL (nameID 12)
    licenseDescription (nameID 13)
    licenseInfoURL (nameID 14)
    typographicFamily (nameID 16)
    typographicSubfamily (nameID 17)
    compatibleFullName (nameID 18)
    sampleText (nameID 19)
    postScriptCIDFindfontName (nameID 20)
    wwsFamilyName (nameID 21)
    wwsSubfamilyName (nameID 22)
    lightBackgroundPalette (nameID 23)
    darkBackgroundPalette (nameID 24)
    variationsPostScriptNamePrefix (nameID 25)
    """
    unique_name = meta_infos.family_name.replace(' ', '-')
    name_strings = {
        'familyName': meta_infos.family_name,
        'styleName': meta_infos.style_name,
        'uniqueFontIdentifier': f'{unique_name}-{meta_infos.style_name};{meta_infos.version}',
        'fullName': f'{meta_infos.family_name} {meta_infos.style_name}',
        'version': meta_infos.version,
        'psName': f'{unique_name}-{meta_infos.style_name}',
    }
    if meta_infos.copyright_info is not None:
        name_strings['copyright'] = meta_infos.copyright_info
    if meta_infos.manufacturer is not None:
        name_strings['manufacturer'] = meta_infos.manufacturer
    if meta_infos.designer is not None:
        name_strings['designer'] = meta_infos.designer
    if meta_infos.description is not None:
        name_strings['description'] = meta_infos.description
    if meta_infos.vendor_url is not None:
        name_strings['vendorURL'] = meta_infos.vendor_url
    if meta_infos.designer_url is not None:
        name_strings['designerURL'] = meta_infos.designer_url
    if meta_infos.license_info is not None:
        name_strings['licenseDescription'] = meta_infos.license_info
    if meta_infos.license_url is not None:
        name_strings['licenseInfoURL'] = meta_infos.license_url
    if meta_infos.sample_text is not None:
        name_strings['sampleText'] = meta_infos.sample_text
    return name_strings


def _create_outlines(glyph_data: list[list[int]], px_to_units: int) -> list[list[tuple[int, int]]]:
    """
    将字形数据转换为轮廓数据，左上角原点坐标系
    """
    # 相邻像素分组
    point_group_list = []
    for y, glyph_data_row in enumerate(glyph_data):
        for x, alpha in enumerate(glyph_data_row):
            if alpha > 0:
                new_point_group = {(x, y)}
                for i, point_group in enumerate(reversed(point_group_list)):
                    # 遍历方向为右下，因此只需检查左上
                    if (x - 1, y) in point_group or (x, y - 1) in point_group:
                        point_group_list.remove(point_group)
                        new_point_group = new_point_group.union(point_group)
                point_group_list.append(new_point_group)
    # 对每组生成轮廓
    outlines = []
    for point_group in point_group_list:
        # 按照像素拆分线段，注意绘制顺序
        pending_line_segments = []
        for (x, y) in point_group:
            point_outline = [
                (x * px_to_units, y * px_to_units),
                ((x + 1) * px_to_units, y * px_to_units),
                ((x + 1) * px_to_units, (y + 1) * px_to_units),
                (x * px_to_units, (y + 1) * px_to_units),
            ]
            # 一个像素有左右上下四条边，如果该边没有相邻像素，则该边线段有效
            if x <= 0 or glyph_data[y][x - 1] <= 0:  # 左
                pending_line_segments.append([point_outline[3], point_outline[0]])
            if x >= len(glyph_data[y]) - 1 or glyph_data[y][x + 1] <= 0:  # 右
                pending_line_segments.append([point_outline[1], point_outline[2]])
            if y <= 0 or glyph_data[y - 1][x] <= 0:  # 上
                pending_line_segments.append([point_outline[0], point_outline[1]])
            if y >= len(glyph_data) - 1 or glyph_data[y + 1][x] <= 0:  # 下
                pending_line_segments.append([point_outline[2], point_outline[3]])
        # 连接所有的线段，注意绘制顺序
        solved_line_segments = []
        while len(pending_line_segments) > 0:
            pending_line_segment = pending_line_segments.pop()
            for i, solved_line_segment in enumerate(reversed(solved_line_segments)):
                left_line_segment = None
                right_line_segment = None
                # 一共4种连接情况
                if pending_line_segment[-1] == solved_line_segment[0]:
                    left_line_segment = pending_line_segment
                    right_line_segment = solved_line_segment
                elif pending_line_segment[-1] == solved_line_segment[-1]:
                    solved_line_segment.reverse()
                    left_line_segment = pending_line_segment
                    right_line_segment = solved_line_segment
                elif solved_line_segment[-1] == pending_line_segment[0]:
                    left_line_segment = solved_line_segment
                    right_line_segment = pending_line_segment
                elif solved_line_segment[-1] == pending_line_segment[-1]:
                    pending_line_segment.reverse()
                    left_line_segment = solved_line_segment
                    right_line_segment = pending_line_segment
                # 需要连接的情况
                if left_line_segment is not None and right_line_segment is not None:
                    solved_line_segments.remove(solved_line_segment)
                    # 连接的两个点是重复的
                    right_line_segment.pop(0)
                    # 判断连接的点是不是可省略
                    x, y = left_line_segment[-1]
                    xl, yl = left_line_segment[-2]
                    xr, yr = right_line_segment[0]
                    if (x == xl and x == xr) or (y == yl and y == yr):
                        left_line_segment.pop()
                    # 连接线段
                    pending_line_segment = left_line_segment + right_line_segment
            solved_line_segments.append(pending_line_segment)
        # 将连接好的线段添加到轮廓数组中，有多条线段的情况，是中间有镂空（绘制顺序与外边框相反）
        for solved_line_segment in solved_line_segments:
            # 首尾的两个点是重复的
            solved_line_segment.pop(0)
            # 判断尾点是不是可省略
            x, y = solved_line_segment[-1]
            xl, yl = solved_line_segment[-2]
            xr, yr = solved_line_segment[0]
            if (x == xl and x == xr) or (y == yl and y == yr):
                solved_line_segment.pop()
            # 添加到轮廓
            outlines.append(solved_line_segment)
    return outlines


def _create_glyph(outlines: list[list[tuple[int, int]]], glyph_context: Glyph, px_to_units: int, is_ttf: bool) -> OTFGlyph | TTFGlyph:
    if is_ttf:
        pen = TTFGlyphPen()
    else:
        pen = OTFGlyphPen(glyph_context.advance_width * px_to_units, None)
    if len(outlines) > 0:
        for outline_index, outline in enumerate(outlines):
            for point_index, point in enumerate(outline):

                # 转换左上角原点坐标系为左下角原点坐标系
                x, y = point
                x += glyph_context.offset_x * px_to_units
                y = (glyph_context.height + glyph_context.offset_y) * px_to_units - y
                point = x, y

                if point_index == 0:
                    pen.moveTo(point)
                else:
                    pen.lineTo(point)
            if outline_index < len(outlines) - 1:
                pen.endPath()
            else:
                pen.closePath()
    else:
        pen.moveTo((0, 0))
        pen.closePath()
    if is_ttf:
        return pen.glyph()
    else:
        return pen.getCharString()


def _get_glyph_with_cache(glyph_context: Glyph, px_to_units: int, is_ttf: bool) -> OTFGlyph | TTFGlyph:
    cache_tag = f'{glyph_context.advance_width}#{glyph_context.offset}#{glyph_context.data}'.replace(' ', '')
    if getattr(glyph_context, _CACHE_NAME_TAG, None) != cache_tag:
        setattr(glyph_context, _CACHE_NAME_OUTLINES, None)
        setattr(glyph_context, _CACHE_NAME_OTF_GLYPH, None)
        setattr(glyph_context, _CACHE_NAME_TTF_GLYPH, None)
        setattr(glyph_context, _CACHE_NAME_TAG, cache_tag)

    outlines = getattr(glyph_context, _CACHE_NAME_OUTLINES, None)
    if outlines is None:
        logger.debug("Create 'Outlines': %s", glyph_context.name)
        outlines = _create_outlines(glyph_context.data, px_to_units)
        setattr(glyph_context, _CACHE_NAME_OUTLINES, outlines)
    else:
        logger.debug("Use cached 'Outlines': %s", glyph_context.name)

    cache_name_xtf_glyph = _CACHE_NAME_TTF_GLYPH if is_ttf else _CACHE_NAME_OTF_GLYPH
    glyph = getattr(glyph_context, cache_name_xtf_glyph, None)
    if glyph is None:
        logger.debug("Create '%sGlyph': %s", 'TTF' if is_ttf else 'OTF', glyph_context.name)
        glyph = _create_glyph(outlines, glyph_context, px_to_units, is_ttf)
        setattr(glyph_context, cache_name_xtf_glyph, glyph)
    else:
        logger.debug("Use cached '%sGlyph': %s", 'TTF' if is_ttf else 'OTF', glyph_context.name)
    return glyph


def create_builder(context: 'font.FontBuilder', is_ttf: bool, flavor: Flavor = None) -> FontBuilder:
    logger.debug("Create '%sBuilder': %s", 'TTF' if is_ttf else 'OTF', context.meta_infos.family_name)
    context.check_ready()

    units_per_em = context.size * context.opentype_configs.px_to_units
    builder = FontBuilder(units_per_em, isTTF=is_ttf)

    logger.debug("Setup 'Name Strings'")
    name_strings = _create_name_strings(context.meta_infos)
    builder.setupNameTable(name_strings)

    logger.debug("Setup 'Glyph Order' and 'Glyphs'")
    glyph_order = ['.notdef']
    glyphs = {}
    for glyph_context in context.get_glyphs():
        if glyph_context.name != '.notdef':
            glyph_order.append(glyph_context.name)
        glyphs[glyph_context.name] = _get_glyph_with_cache(glyph_context, context.opentype_configs.px_to_units, is_ttf)
    builder.setupGlyphOrder(glyph_order)
    if is_ttf:
        builder.setupGlyf(glyphs)
    else:
        if context.opentype_configs.cff_family_name is not None:
            cff_family_name = context.opentype_configs.cff_family_name
        else:
            cff_family_name = context.meta_infos.family_name
        cff_ps_name = f'{cff_family_name.replace(" ", "-")}-{context.meta_infos.style_name}'
        cff_font_infos = {
            'FamilyName': cff_family_name,
            'FullName': f'{cff_family_name} {context.meta_infos.style_name}',
            'Weight': context.meta_infos.style_name,
        }
        if context.meta_infos.copyright_info is not None:
            cff_font_infos['Notice'] = context.meta_infos.copyright_info
        builder.setupCFF(cff_ps_name, cff_font_infos, glyphs, {})

    logger.debug("Setup 'Character Mapping'")
    builder.setupCharacterMap(context.character_mapping)

    logger.debug("Setup 'Horizontal Metrics'")
    horizontal_metrics = {}
    for glyph_name in glyph_order:
        advance_width = context.get_glyph(glyph_name).advance_width * context.opentype_configs.px_to_units
        if is_ttf:
            lsb = glyphs[glyph_name].xMin
        else:
            lsb = glyphs[glyph_name].calcBounds(None)[0]
        horizontal_metrics[glyph_name] = advance_width, lsb
    builder.setupHorizontalMetrics(horizontal_metrics)

    ascent = context.ascent * context.opentype_configs.px_to_units
    descent = context.descent * context.opentype_configs.px_to_units
    x_height = context.x_height * context.opentype_configs.px_to_units if context.x_height is not None else None
    cap_height = context.cap_height * context.opentype_configs.px_to_units if context.cap_height is not None else None
    logger.debug("Setup 'Horizontal Header'")
    builder.setupHorizontalHeader(
        ascent=ascent,
        descent=descent,
    )
    logger.debug("Setup 'OS2'")
    builder.setupOS2(
        sTypoAscender=ascent,
        sTypoDescender=descent,
        usWinAscent=ascent,
        usWinDescent=-descent,
        sxHeight=x_height,
        sCapHeight=cap_height,
    )

    logger.debug("Setup 'Post'")
    builder.setupPost()

    if flavor is not None:
        logger.debug("Setup 'Flavor': %s", flavor)
        builder.font.flavor = flavor

    logger.debug("Create '%sBuilder' finished", 'TTF' if is_ttf else 'OTF')
    return builder


def create_collection_builder(context: 'font.FontCollectionBuilder', is_ttf: bool) -> TTCollection:
    collection_builder = TTCollection()
    for font_context in context.font_builders:
        collection_builder.fonts.append(create_builder(font_context, is_ttf).font)
    return collection_builder



import os

import bdffont
import fontTools.fontBuilder
import fontTools.ttLib

from pixel_font_builder import opentype, bdf
from pixel_font_builder.glyph import Glyph
from pixel_font_builder.info import MetaInfos


class FontBuilder:
    def __init__(
            self,
            size: int,
            ascent: int,
            descent: int,
            x_height: int = None,
            cap_height: int = None,
    ):
        self.size = size
        self.ascent = ascent
        self.descent = descent
        self.x_height = x_height
        self.cap_height = cap_height

        self.character_mapping = dict[int, str]()
        self._name_to_glyph = dict[str, Glyph]()

        self.meta_infos = MetaInfos()
        self.opentype_configs = opentype.Configs()
        self.bdf_configs = bdf.Configs()

    @property
    def line_height(self) -> int:
        return self.ascent - self.descent

    def get_glyph(self, name: str) -> Glyph | None:
        return self._name_to_glyph.get(name, None)

    def get_glyphs(self) -> list[Glyph]:
        return list(self._name_to_glyph.values())

    def get_glyphs_count(self) -> int:
        return len(self._name_to_glyph)

    def set_glyph(self, glyph: Glyph):
        self._name_to_glyph[glyph.name] = glyph

    def add_glyph(self, glyph: Glyph):
        if glyph.name in self._name_to_glyph:
            raise Exception(f"Glyph '{glyph.name}' already exists")
        self._name_to_glyph[glyph.name] = glyph

    def add_glyphs(self, glyphs: list[Glyph]):
        for glyph in glyphs:
            self.add_glyph(glyph)

    def remove_glyph(self, name: str) -> Glyph | None:
        return self._name_to_glyph.pop(name, None)

    def check_ready(self):
        if '.notdef' not in self._name_to_glyph:
            raise Exception("Need to provide a glyph named '.notdef'")
        for code_point, glyph_name in self.character_mapping.items():
            if code_point < 0:
                raise Exception('Code points must >= 0')
            if glyph_name not in self._name_to_glyph:
                raise Exception(f"Missing glyph: '{glyph_name}'")
        for glyph in self._name_to_glyph.values():
            glyph.check_ready()
        if self.meta_infos.version is None:
            raise Exception("Missing meta infos: 'version'")
        if self.meta_infos.family_name is None:
            raise Exception("Missing meta infos: 'family_name'")
        if self.meta_infos.style_name is None:
            raise Exception("Missing meta infos: 'style_name'")

    def to_otf_builder(self, flavor: opentype.Flavor = None) -> fontTools.fontBuilder.FontBuilder:
        return opentype.create_builder(self, False, flavor)

    def save_otf(
            self,
            file_path: str | bytes | os.PathLike[str] | os.PathLike[bytes],
            flavor: opentype.Flavor = None,
    ):
        self.to_otf_builder(flavor).save(file_path)

    def to_ttf_builder(self, flavor: opentype.Flavor = None) -> fontTools.fontBuilder.FontBuilder:
        return opentype.create_builder(self, True, flavor)

    def save_ttf(
            self,
            file_path: str | bytes | os.PathLike[str] | os.PathLike[bytes],
            flavor: opentype.Flavor = None,
    ):
        self.to_ttf_builder(flavor).save(file_path)

    def to_bdf_builder(self) -> bdffont.BdfFont:
        return bdf.create_builder(self)

    def save_bdf(
            self,
            file_path: str | bytes | os.PathLike[str] | os.PathLike[bytes],
            optimize_bitmap: bool = True,
    ):
        self.to_bdf_builder().save(file_path, optimize_bitmap)


class FontCollectionBuilder:
    def __init__(self, font_builders: list[FontBuilder] = None):
        if font_builders is None:
            font_builders = list[FontBuilder]()
        self.font_builders = font_builders

    def to_otc_builder(self) -> fontTools.ttLib.TTCollection:
        return opentype.create_collection_builder(self, False)

    def save_otc(
            self,
            file_path: str | bytes | os.PathLike[str] | os.PathLike[bytes],
            share_tables: bool = True,
    ):
        self.to_otc_builder().save(file_path, share_tables)

    def to_ttc_builder(self) -> fontTools.ttLib.TTCollection:
        return opentype.create_collection_builder(self, True)

    def save_ttc(
            self,
            file_path: str | bytes | os.PathLike[str] | os.PathLike[bytes],
            share_tables: bool = True,
    ):
        self.to_ttc_builder().save(file_path, share_tables)
